@@ -15,18 +15,10 @@ var differ = function(beforeText, afterText, userParams) {
   var sm = new difflib.SequenceMatcher(this.beforeLines, this.afterLines);
   this.opcodes = sm.get_opcodes();
 
-  this.suppressCharDiffs_ = !(
-      differ.suitableForHighlighting_(this.beforeLines) &&
-      differ.suitableForHighlighting_(this.afterLines));
-
   if (this.params.language) {
-    if (!this.suppressCharDiffs_) {
-      var lang = this.params.language;
-      this.beforeLinesHighlighted = differ.highlightText_(beforeText, lang);
-      this.afterLinesHighlighted = differ.highlightText_(afterText, lang);
-    } else {
-      this.params.language = null;
-    }
+    var lang = this.params.language;
+    this.beforeLinesHighlighted = differ.highlightText_(beforeText, lang);
+    this.afterLinesHighlighted = differ.highlightText_(afterText, lang);
   }
 };
 
@@ -34,20 +26,6 @@ differ.prototype.maxLineNumber = function() {
   return Math.max(this.beforeLines.length, this.afterLines.length);
 };
 
-/**
- * Check for long lines which might make highlight.js hang.
- * @param {Array.string>} lines The lines which might be highlighted.
- */
-differ.suitableForHighlighting_ = function(lines) {
-  var MAX_LINE_LENGTH = 1000;
-  for (var i = 0; i < lines.length; i++) {
-    if (lines[i].length > MAX_LINE_LENGTH) {
-      return false;
-    }
-  }
-  return true;
-};
-
 /**
  * @param {string} text Possibly multiline text containing spans that cross
  *     line breaks.
@@ -171,7 +149,7 @@ differ.prototype.buildRow_ = function(beforeIdx, beforeEnd, afterIdx, afterEnd,
   beforeIdx = addCells(els, beforeIdx, beforeEnd, this.params.language, beforeLines, 'before ' + change, beforeIdx + 1);
   afterIdx = addCells(els, afterIdx, afterEnd, this.params.language, afterLines, 'after ' + change, afterIdx + 1);
 
-  if (change == 'replace' && !this.suppressCharDiffs_) {
+  if (change == 'replace') {
     differ.addCharacterDiffs_(els[1], els[3], this.params.language);
   }
 
@@ -182,6 +160,65 @@ differ.prototype.buildRow_ = function(beforeIdx, beforeEnd, afterIdx, afterEnd,
   };
 };
 
+// Construct ranges of lines to show consecutively on either side.
+// The main work of this is factoring out ranges of common lines.
+// Output:
+// [ {change: "equal", left: { start, end }, right: { start, end }}, ... ]
+differ.prototype.buildRanges_ = function() {
+  var contextSize = this.params.contextSize;
+  var ranges = [];
+
+  for (var opcodeIdx = 0; opcodeIdx < this.opcodes.length; opcodeIdx++) {
+    var opcode = this.opcodes[opcodeIdx];
+    var change = opcode[0];  // "equal", "replace", "delete", "insert"
+    var beforeIdx = opcode[1];
+    var beforeEnd = opcode[2];
+    var afterIdx = opcode[3];
+    var afterEnd = opcode[4];
+    var rowCount = Math.max(beforeEnd - beforeIdx, afterEnd - afterIdx);
+    var topRows = [];
+    var botRows = [];
+
+    for (var i = 0; i < rowCount; i++) {
+      // Jump
+      if (contextSize && this.opcodes.length > 1 && change == 'equal' &&
+          ((opcodeIdx > 0 && i == contextSize) ||
+           (opcodeIdx == 0 && i == 0))) {
+        var jump = rowCount - ((opcodeIdx == 0 ? 1 : 2) * contextSize);
+        var isEnd = (opcodeIdx + 1 == this.opcodes.length);
+        if (isEnd) {
+          jump += (contextSize - 1);
+        }
+        if (jump > 1) {
+          ranges.push({
+            type: 'skip',
+            left: { start: beforeIdx, end: beforeIdx + jump },
+            right: { start: afterIdx, end: afterIdx + jump }
+          });
+
+          beforeIdx += jump;
+          afterIdx += jump;
+          i += jump - 1;
+
+          // skip last lines if they're all equal
+          if (isEnd) {
+            break;
+          } else {
+            continue;
+          }
+        }
+      }
+
+      var data = this.buildRow_(beforeIdx, beforeEnd, afterIdx, afterEnd, change);
+      beforeIdx = data.newBeforeIdx;
+      afterIdx = data.newAfterIdx;
+      topRows.push(data.row);
+    }
+  }
+
+  return ranges;
+};
+
 differ.prototype.buildView_ = function() {
   var $leftLineDiv = $('<div class="diff-line-no diff-left diff-left-line-no">');
   var $leftContent = $('<div class="diff-content diff-left-content">');
@@ -222,7 +259,7 @@ differ.prototype.buildView_ = function() {
             'afterStartIndex': afterIdx,
             'jumpLength': jump,
           }).attr('line-no', 1 + afterIdx);
-          
+
           els.push($('<div class=line-no>&hellip;</div>').attr('line-no', 1+beforeIdx).get(0));
           els.push($('<div class="skip code">...</div>').attr('line-no', 1+beforeIdx).get(0));
           els.push($('<div class=line-no>&hellip;</div>').attr('line-no', 1+afterIdx).get(0));
@@ -231,7 +268,7 @@ differ.prototype.buildView_ = function() {
           beforeIdx += jump;
           afterIdx += jump;
           i += jump - 1;
-          
+
           // skip last lines if they're all equal
           if (isEnd) {
             break;
@@ -271,7 +308,6 @@ differ.prototype.buildView_ = function() {
       )
       );
 
-  // TODO(danvk): append each element of rows to the appropriate div here.
   rows.forEach(function(row) {
     if (row.length != 4) throw "Invalid row: " + row;
 
@@ -329,9 +365,9 @@ function html_substr(html, start, count) {
   var div = document.createElement('div');
   div.innerHTML = html;
   var consumed = 0;
-  
+
   walk(div, track);
-  
+
   function track(el) {
     if (count > 0) {
       var len = el.data.length;
@@ -352,7 +388,7 @@ function html_substr(html, start, count) {
       el.data = '';
     }
   }
-  
+
   function walk(el, fn) {
     var node = el.firstChild, oldNode;
     var elsToRemove = [];
@@ -385,37 +421,98 @@ function html_substr(html, start, count) {
   return div.innerHTML;
 }
 
+/**
+ * @param {string} line The line to be split
+ * @return {Array.<string>} Component words in the line. An invariant is that
+ *     splitIntoWords_(line).join('') == line.
+ */
+differ.splitIntoWords_ = function(line) {
+  var LC = 0, UC = 2, NUM = 3, WS = 4, SYM = 5;
+  var charType = function(c) {
+    if (c.match(/[a-z]/)) return LC;
+    if (c.match(/[A-Z]/)) return UC;
+    if (c.match(/[0-9]/)) return NUM;
+    if (c.match(/\s/)) return WS;
+    return SYM;
+  };
 
-differ.addCharacterDiffs_ = function(beforeCell, afterCell) {
-  var beforeText = $(beforeCell).text(),
-      afterText = $(afterCell).text(),
-      beforeHtml = $(beforeCell).html(),
-      afterHtml = $(afterCell).html();
-  var sm = new difflib.SequenceMatcher(beforeText.split(''), afterText.split(''));
+  // TODO: consider putting each whitespace char into its own word.
+  // Single words can be [A-Z][a-z]+, [A-Z]+, [a-z]+, [0-9]+ or \s+.
+  var words = [];
+  var lastType = -1;
+  for (var i = 0; i < line.length; i++) {
+    var c = line.charAt(i);
+    var ct = charType(c);
+    if (ct == lastType && ct != SYM ||
+        ct == LC && lastType == UC && words[words.length - 1].length == 1) {
+      words[words.length - 1] += c;
+    } else {
+      words.push(c);
+    }
+    lastType = ct;
+  }
+  return words;
+};
+
+/**
+ * Compute an intra-line diff.
+ * @param {string} beforeText
+ * @param {string} afterText
+ * @return {?Array.<Array>} [before codes, after codes], where each element is a
+ *     list of ('change type', start idx, stop idx) triples. Returns null if
+ *     character differences are not appropriate for this line pairing.
+ */
+differ.computeCharacterDiffs_ = function(beforeText, afterText) {
+  var beforeWords = differ.splitIntoWords_(beforeText),
+      afterWords = differ.splitIntoWords_(afterText);
+
+  // TODO: precompute two arrays; this does too much work.
+  var wordToIdx = function(isBefore, idx) {
+    var words = isBefore ? beforeWords : afterWords;
+    var charIdx = 0;
+    for (var i = 0; i < idx; i++) {
+      charIdx += words[i].length;
+    }
+    return charIdx;
+  };
+
+  var sm = new difflib.SequenceMatcher(beforeWords, afterWords);
   var opcodes = sm.get_opcodes();
-  var minEqualFrac = 0.5;  // suppress character-by-character diffs if there's less than this much overlap.
+
+  // Suppress char-by-char diffs if there's less than 50% character overlap.
+  // The one exception is pure whitespace diffs, which should always be shown.
+  var minEqualFrac = 0.5;
   var equalCount = 0, charCount = 0;
+  var beforeDiff = '', afterDiff = '';
   opcodes.forEach(function(opcode) {
     var change = opcode[0];
-    var beforeLen = opcode[2] - opcode[1];
-    var afterLen = opcode[4] - opcode[3];
+    var beforeIdx = wordToIdx(true, opcode[1]);
+    var beforeEnd = wordToIdx(true, opcode[2]);
+    var afterIdx = wordToIdx(false, opcode[3]);
+    var afterEnd = wordToIdx(false, opcode[4]);
+    var beforeLen = beforeEnd - beforeIdx;
+    var afterLen = afterEnd - afterIdx;
     var count = beforeLen + afterLen;
-    if (change == 'equal') equalCount += count;
+    if (change == 'equal') {
+      equalCount += count;
+    } else {
+      beforeDiff += beforeText.substring(beforeIdx, beforeEnd);
+      afterDiff += afterText.substring(afterIdx, afterEnd);
+    }
     charCount += count;
   });
-  if (equalCount < minEqualFrac * charCount) return;
-
-  var m = differ.htmlTextMapper.prototype.getHtmlSubstring;
-  var beforeMapper = new differ.htmlTextMapper(beforeText, beforeHtml);
-  var afterMapper = new differ.htmlTextMapper(afterText, afterHtml);
+  if (equalCount < minEqualFrac * charCount &&
+      !(beforeDiff.match(/^\s*$/) && afterDiff.match(/^\s*$/))) {
+    return null;
+  }
 
   var beforeOut = [], afterOut = [];  // (span class, start, end) triples
   opcodes.forEach(function(opcode) {
     var change = opcode[0];
-    var beforeIdx = opcode[1];
-    var beforeEnd = opcode[2];
-    var afterIdx = opcode[3];
-    var afterEnd = opcode[4];
+    var beforeIdx = wordToIdx(true, opcode[1]);
+    var beforeEnd = wordToIdx(true, opcode[2]);
+    var afterIdx = wordToIdx(false, opcode[3]);
+    var afterEnd = wordToIdx(false, opcode[4]);
     if (change == 'equal') {
       beforeOut.push([null, beforeIdx, beforeEnd]);
       afterOut.push([null, afterIdx, afterEnd]);
@@ -433,6 +530,29 @@ differ.addCharacterDiffs_ = function(beforeCell, afterCell) {
   beforeOut = differ.simplifyCodes_(beforeOut);
   afterOut = differ.simplifyCodes_(afterOut);
 
+  return [beforeOut, afterOut];
+};
+
+
+// Add character-by-character diffs to a row (if appropriate).
+differ.addCharacterDiffs_ = function(beforeCell, afterCell) {
+  var beforeText = $(beforeCell).text(),
+      afterText = $(afterCell).text();
+  var codes = differ.computeCharacterDiffs_(beforeText, afterText);
+  if (codes == null) return;
+  beforeOut = codes[0];
+  afterOut = codes[1];
+
+  // Splice in "insert", "delete" and "replace" tags.
+  // This is made more difficult by the presence of syntax highlighting, which
+  // has its own set of tags. The two can co-exists if we're careful to only
+  // wrap complete (balanced) DOM trees.
+  var beforeHtml = $(beforeCell).html(),
+      afterHtml = $(afterCell).html();
+  var m = differ.htmlTextMapper.prototype.getHtmlSubstring;
+  var beforeMapper = new differ.htmlTextMapper(beforeText, beforeHtml);
+  var afterMapper = new differ.htmlTextMapper(afterText, afterHtml);
+
   $(beforeCell).empty().html(differ.codesToHtml_(beforeMapper, beforeOut));
   $(afterCell).empty().html(differ.codesToHtml_(afterMapper, afterOut));
 };
